#ifndef FUNCTION_CACHE_HPP
#define FUNCTION_CACHE_HPP

#include <type_traits>
#include <map>
#include <tuple>

auto add_cache = [](auto fun)
{
    using fun_type = decltype(fun);
    return ([=](auto... run_args)
    {
        using fun_return_type = std::result_of_t<fun_type(decltype(run_args)...)>;
        static std::map<
            std::tuple<fun_type, decltype(run_args)...>,
            fun_return_type
        > result_cache;
        std::tuple<fun_type, decltype(run_args)...> tuple(fun, run_args...);
        auto search = result_cache.find(tuple);
        if ( search == result_cache.end())
        {
            fun_return_type rv = fun(run_args...);
            result_cache[tuple] = rv;
            return rv;
        }
        else
        {
            return search->second;
        }
    });
};

#endif
